
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#include <sys/types.h>                      /* time_t */
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <stdio.h>                          /* printf() */
#include <assert.h>
#include <stdarg.h>
#include <syslog.h>

#include "initng_syslog.h"
#include "../../src/initng_active_status.h"
#include "../../src/initng_global.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"

static void syslog_fetch_output(active_h * service, int datalen, process_h * process);
static void syslog_print_system_state(h_sys_state state);
static int syslog_print_status_change(active_h * service);
static void check_syslog(void);
static void initng_log(int prio, const char *owner, const char *format, ...);
static void free_buffert(void);

/* global variable */
int syslog_running;

static void check_syslog(void)
{
    struct stat st;
    
    if(stat("/dev/log", &st)==0 && S_ISSOCK(st.st_mode))
    {
    	syslog_running=1;
	
	/* print out the buffert if any now */
	if(!list_empty(&log_list.list))
	{
	    log_ent *current, *safe=NULL;
	    while_log_list_safe(current, safe)
	    {
		initng_log(current->prio, current->owner, "%s", current->buffert);
		free(current->buffert);
		if(current->owner)
		    free(current->owner);
		list_del(&current->list);
	    }
	    INIT_LIST_HEAD(&log_list.list);
	}
    }
    else {
	syslog_running=0;
    }
}

static void free_buffert(void)
{
    log_ent *current, *safe=NULL;
    
    /* give syslog a last chanche to come alive */
    check_syslog();
    while_log_list_safe(current, safe)
    {
	free(current->buffert);
	if(current->owner)
	    free(current->owner);
	list_del(&current->list);
    }
    INIT_LIST_HEAD(&log_list.list);
    
    /* log directly to syslog from now, even if it might not exist */
    syslog_running=1;
}

static void initng_log(int prio, const char *owner, const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    
	
    /* if syslog is running, send it directly */
    if(syslog_running==1)
    {
	/* if owner is set, we have openlog with right owner */
	if(owner)
	{
	    closelog();
	    openlog (owner, 0, LOG_LOCAL1);
	    
	    vsyslog(prio, format, ap);
	    
	    closelog();
	    openlog ("InitNG", 0, LOG_LOCAL1);
	}
	else
	    vsyslog(prio, format, ap);

    } else {
	char b[201];
	log_ent *tmp=(log_ent *)initng_calloc(1, sizeof(log_ent));
	if(!tmp) return;

	tmp->prio=prio;
	vsnprintf(b, 200, format, ap);
	tmp->buffert=i_strdup(b);
	
	if(owner)
	    tmp->owner=i_strdup(owner);
	else
	    tmp->owner=NULL;
	
	list_add(&tmp->list, &log_list.list);
	    
    }
    
    va_end(ap);
}

/* add values to syslog database */
static int syslog_print_status_change(active_h * service)
{
    assert(service);
    assert(service->name);
    switch(service->a_status)
    {
	case DONE:
	case RUNNING:
	    check_syslog();
	    initng_log (LOG_NOTICE, NULL, "Service %s is started.\n", service->name);
	    break;
	case STOPPED:
	    initng_log (LOG_NOTICE, NULL, "Service %s has stopped.\n", service->name);
	    break;
	case FAIL_STARTING:
	    initng_log (LOG_NOTICE, NULL, "Service %s FAILED to start.\n", service->name);
	    break;
	case FAIL_STOPPING:
	    initng_log (LOG_NOTICE, NULL, "Service %s FAILED to stop.\n", service->name);
	    break;
	default:
	    break;
    }
    /* leave */
    return (TRUE);
}

static void syslog_print_system_state(h_sys_state state)
{
    switch(state)
    {
	case STATE_UP:
	    initng_log (LOG_NOTICE, NULL, "System is up and running!\n");
	    /*
	     * if syslogd have not been started yet, its no ida,
	     * to spend memory for the buffert anyway.
	     */
	    free_buffert();
	    break;
	case STATE_STARTING:
	    initng_log (LOG_NOTICE, NULL, "System is starting up.\n");
	    break;
	case STATE_STOPPING:
	    initng_log (LOG_NOTICE, NULL, "System is going down.\n");
	    break;
	case STATE_ASE:
	    initng_log (LOG_NOTICE, NULL, "Last service exited.\n");
	    break;
	case STATE_EXIT:
	    initng_log (LOG_NOTICE, NULL, "Initng is exiting.\n");
	    break;
	case STATE_RESTART:
	    initng_log (LOG_NOTICE, NULL, "Initng is restarting.\n");
	    break;
	case STATE_HALT:
	    initng_log (LOG_NOTICE, NULL, "System is halting.\n");
	    break;
	case STATE_REBOOT:
	    initng_log (LOG_NOTICE, NULL, "System is rebooting.\n");
	    break;
	default:
	    break;
    }

}

static void syslog_fetch_output(active_h * service, int datalen, process_h * process)
{
    char log[200];
    int pos = process->buffer_pos;
    int i;
    
    assert(service);
    assert(service->name);
    
    /* print every line, ending with a '\n' as an own syslog */
    while( process->buffer[pos] )
    {
	i=0;
	/* count the number of char before '\n' */
	while(process->buffer[pos+i] && process->buffer[pos+i]!='\n' && i<200)
	    i++;
	    
	/* copy that many chars to our temporary log array */
	strncpy(log, &process->buffer[pos], i);
	log[i]='\0';

	/* send it to syslog */
	initng_log (LOG_NOTICE, service->name, "%s", log);
	
	/* step forward */
	pos+=i;
	
	/* and skip the newline if any */
	if(process->buffer[pos])
	    pos++;
	    
    }
}


static int syslog_print_error(e_mt mt, const char *file, const char *func, int line,
                const char *format, va_list arg)
{
    assert(file);
    assert(func);

    switch (mt)
    {
        case MSG_FAIL:
            syslog(LOG_EMERG, "\"%s\", %s() #%i FAIL: ", file, func, line);
	    vsyslog(LOG_EMERG, format, arg);
            break;
        case MSG_WARN:
            syslog(LOG_WARNING, "\"%s\", %s() #%i WARN: ", file, func, line);
	    vsyslog(LOG_WARNING, format, arg);
            break;
	default:
            syslog(LOG_NOTICE, "\"%s\", %s() #%i : ", file, func, line);
	    vsyslog(LOG_NOTICE, format, arg);
            break;
    }


    return (TRUE);
}




int module_init(const char *version)
{
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module, is compiled for \"%s\" version, and initng is compiled on \"%s\" version, wont load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    INIT_LIST_HEAD(&log_list.list);
    check_syslog();

    setlogmask (LOG_UPTO (LOG_NOTICE));
    openlog ("InitNG", 0 , LOG_LOCAL1);
    
    initng_add_hook(ERR_MSG, 50, &syslog_print_error);
    initng_add_hook(ASTATUS_CHANGE, 100, &syslog_print_status_change);
    initng_add_hook(SWATCHERS, 100, &syslog_print_system_state);
    initng_add_hook(PIPEWATCHERS, 100, &syslog_fetch_output);

    return (TRUE);
}


void module_unload(void)
{
    initng_del_hook(ASTATUS_CHANGE, &syslog_print_status_change);
    initng_del_hook(SWATCHERS, &syslog_print_system_state);
    initng_del_hook(PIPEWATCHERS, &syslog_fetch_output);
    initng_del_hook(ERR_MSG, &syslog_print_error);
    free_buffert();
    closelog();
}
