/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <dlfcn.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>                          /* fcntl() */
#include "initng.h"
#include "initng_service_cache.h"
#include "initng_load_module.h"
#include "initng_toolbox.h"
#include "initng_global.h"
#include "initng_signal.h"

#include "initng_plugin_callers.h"
#include "initng_plugin_type.h"

#include "initng_fd.h"

void initng_close_all_fd(void)
{
    s_call *current = NULL;

    S_;

    while_call_db(current)
    {
        if (current->ptype != FDWATCHERS) continue;
        
        close(current->c.fdh->fds);
        list_del(&current->list);

        if (current->from_file)
            free(current->from_file);
        free(current);
    }
}

/* Calls plugins when data is ready in pipe */
void initng_plugin_readpipe(active_h * service, int datalen,
                            process_h * process, int force_flush)
{
    s_call *current = NULL;

    S_;

    while_call_db(current)
    {
        if (current->ptype != PIPEWATCHERS) continue;
        
        D_("Calling pipewatcher plugin.\n");
        (*current->c.pipewatcher) (service, datalen, process, force_flush);
    }
}

/* This does the actual read from pipe, all checks have been done here */
/* This code has ben written by neuron, of course its bad commented */
int initng_process_readpipe_read(active_h * service, process_h * x,
                                 int program_end)
{
    int forceflush, maxread, len;

    S_;

    forceflush = program_end;
    maxread = MAX_BUFFER - x->buffer_pos;

    do
    {
        errno = 0;
        len = read(x->stdout[0], &x->buffer[x->buffer_pos], maxread);
    }
    while (errno == EINTR);

    if (len > 0)
    {
        x->buffer[x->buffer_pos + len] = '\0';

        if (x->buffer_pos + len >= MAX_BUFFER)
            forceflush = TRUE;

        initng_plugin_readpipe(service, len, x, forceflush);
        x->buffer_pos = x->buffer_pos + len;
        if (forceflush)
            x->buffer_pos = 0;
    }
    return len;
}

/* This does all the fancy pancy around the read */
void initng_process_readpipe(active_h * service, process_h * x,
                             int program_end)
{
    int i;

    S_;

    /* We've already ended */
    if ((!x->buffer))
        return;


    /*
     * program_end is set, and radpipe is called just
     * before process are beeing freed, when the pocess
     * is dead.
     */

    if (program_end)
    {
        int fd_flags;

        fd_flags = fcntl(x->stdout[0], F_GETFL, 0);
        fcntl(x->stdout[0], F_SETFL, fd_flags | O_NONBLOCK);
    }

    i = initng_process_readpipe_read(service, x, program_end);

    if (!program_end)
    {
        /* We have an error other than EAGAIN, set program_end */
        if ((i == -1) && (errno != EAGAIN))
        {
            program_end = TRUE;
            F_("Error reading from pipe for %s (%s)\n", service->name,
               strerror(errno));
        }
        /* End of file */
        if (i == 0)
            program_end = TRUE;
    }

    /* If i < 1 we either have 0 (end of file) or -1 (error) */
    if ( !program_end)  return;
    
    /* Program is ending, close pipes */
    initng_plugin_readpipe(service, 0, x, TRUE);

    /* free data, close and reset fds */
    D_("Freeing buffer for %s\n", service->name);
    if (x->stdout[0] > 0)
        close(x->stdout[0]);
    if (x->stdout[1] > 0)
        close(x->stdout[1]);
    if (x->buffer)
        free(x->buffer);
    x->stdout[0] = -1;
    x->stdout[1] = -1;
    x->buffer = NULL;
}


/*
 * FILEDESCRIPTORPOLLNG
 *
 * Filedescriptors are scaned in 3 steps.
 * First, scan all plugins, and processes after filedescriptors to
 * watch, and add them with FD_SET(); added++;
 *
 * Second, Run select if there where any added, with an timeout set
 * when function was called.
 *
 * Third, if select returns that there is data to fetch, walk over
 * plugins and filedescriptors again, and call plugins or initng_process_readpipe
 * with pointer to plugin/service.
 *
 * This function will return FALSE, if it timeout, or if there was
 * no fds to monitor.
 */
int initng_plugin_fd_poll(int timeout)
{
    /* Variables */
    fd_set readset;
    struct timeval tv;
    int retval;
    int added = 0;
    active_h *currentA, *qA = NULL;
    s_call *current, *q = NULL;

    /* initziation */
    S_;
    FD_ZERO(&readset);

    /* Set timeval struct, to our timeout */
    tv.tv_sec = timeout;
    tv.tv_usec = 0;


    /*
     * STEP 1:  Scan for fds to add
     */


    /* scan through active plugins that have listening file descriptors, and add them */
    while_call_db_safe(current, q)
    {
	/* first, some checking */
        if (current->ptype != FDWATCHERS) continue;
	if (current->c.fdh->fds <= 2) continue;
	if (!current->c.fdh->call_module) continue;
        
        /*D_("adding fd #%i, from an call_db.\n", current->c.fdh->fds); */
        FD_SET(current->c.fdh->fds, &readset);
        added++;
    }

    /* Then add all file descriptors from process read pipes */
    while_active_db_safe(currentA, qA)
    {
        if ((currentA->start_process)
            && (currentA->start_process->stdout[0] > 2))
        {
            /*D_("adding fd #%i, from %s->start_process->stdout[0]\n", currentA->start_process->stdout[0], currentA->name); */
            FD_SET(currentA->start_process->stdout[0], &readset);
            added++;
        }
        if ((currentA->stop_process)
            && (currentA->stop_process->stdout[0] > 2))
        {
            /*D_("adding fd #%i, from %s->stop_process->stdout[0]\n", currentA->stop_process->stdout[0], currentA->name); */
            FD_SET(currentA->stop_process->stdout[0], &readset);
            added++;
        }
    }


    /*
     * STEP 2: Do the select-poll, if any fds where added
     */

    /* check if there are any set */
    if (added <= 0)
    {
        D_("No filedescriptors set.\n");
        sleep(timeout);
        return (TRUE);
    }
    D_("%i filedescriptors added.\n", added);


    /* If interrupt then return and do main loop, it doesn't hurt if the filebuffers have a bit more data anyway */
    if (g.interrupt)
        return (TRUE);

    /* make the select */
    retval = select(256, &readset, NULL, NULL, &tv);

    /* if interupt is set, dont continue handeling fussy fds */
    if (g.interrupt)
        return (TRUE);

    /* error */
    if (retval < 0)
    {
        F_("Select returned with %d %s\n", retval, strerror(errno));
        return (FALSE);
    }

    /* none was found */
    if (retval == 0)
    {
        D_("There was no data found on any added fd.\n");
        return (FALSE);
    }

    D_("%d fd's active\n", retval);

    /* If a fsck is running select will always return one filehandler active, we give it 0.1 seconds to get some more data into the buffer. */
    sleep(0.1);


    /*
     * STEP 3:  Find the plugins/processes that handles the fd input, and call them
     */

    /* Now scan throught callers */
    current = q = NULL;
    while_call_db_safe(current, q)
    {
        if (current->ptype != FDWATCHERS) continue;
	if (current->c.fdh->fds <= 2) continue;
	if (!current->c.fdh->call_module) continue;
	if (!FD_ISSET(current->c.fdh->fds, &readset)) continue;
        
        D_("Calling pluginhandler for fd #%i\n", current->c.fdh->fds);
        (*current->c.fdh->call_module) ();
        D_("pluginhandler for fd #%i returned!\n", current->c.fdh->fds);

        /* Found match, that means we need to look for one less, if we've found all we should then return */
        retval--;
        if (retval == 0)
            return (TRUE);
    }

    /* Now scan through processes */
    currentA = qA = NULL;
    while_active_db_safe(currentA, qA)
    {
        if (currentA->start_process && currentA->start_process->stdout[0] > 2)
	{
    	    /*D_("Checking %s->start_process on fd %i\n", currentA->name, currentA->start_process->stdout[0]); */
    	    if (FD_ISSET(currentA->start_process->stdout[0], &readset))
            {
                D_("Will read from %s->start_process on fd #%i\n",
                   currentA->name, currentA->start_process->stdout[0]);
                /* Do the actual read from pipe */
                initng_process_readpipe(currentA, currentA->start_process,
                                        FALSE);
                /* If we've been interrupted, return to main loop (program might have ended) */
                if (g.interrupt)
                    return (TRUE);
                /* Found match, that means we need to look for one less, if we've found all we should then return */
                retval--;
                if (retval == 0)
                    return (TRUE);
            }
        }
        if (currentA->stop_process && currentA->stop_process->stdout[0] > 2)
        {
            /*D_("Checking %s-stop_process on fd %i\n", currentA->name, currentA->stop_process->stdout[0]); */
            if (FD_ISSET(currentA->stop_process->stdout[0], &readset))
            {
                D_("Will read from %s->stop_process on fd #%i\n",
                   currentA->name, currentA->stop_process->stdout[0]);
                /* See above */
                initng_process_readpipe(currentA, currentA->stop_process,
                                        FALSE);
                if (g.interrupt)
                    return (TRUE);
                retval--;
                if (retval == 0)
                    return (TRUE);
            }
        }
    }

    return (TRUE);
}
